from __future__ import print_function
import sys

# Alias FileNotFoundError for python 2
try: 
	FileNotFoundError
except NameError:
	FileNotFoundError = IOError	

# Alias/shortcut for error printing
def printerr(*args, **kwargs):
	print(*args, file=sys.stderr, **kwargs)


import random
import re
from itertools import repeat

class Pattern:
	all_sigs = {'d', 's', 'w', 'W', 'c'}
	all_flags = {'~', '=', '+', '^'}
	expression_re = re.compile((	r'%(?:'
										+r'(?P<sig>{0})(?P<flags>{1}*)'
										+r'|\{{(?P<sigm>{0}+)(?P<flagsm>{1}*)\}}'
									+r')'
									+r'(?:\['
										+r'(?P<length_lower>\d+)(?:-(?P<length_upper>\d+))?'
									+r'\])?'
								).format(	
									'[{}]'.format(''.join(all_sigs)),
									'[{}]'.format(''.join('\\'+f for f in all_flags))
								))
	pools_dict = {
		'd': {str(n) for n in range(0,10)},
		's': {'!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '[', ']', '-', '+', '<', '>', '/', '?'},
		'w': {chr(ord('a')+i) for i in range(0,26)},
	}
	pools_dict['c'] = {c for _, pool in pools_dict.items() for c in pool}

	class Expression:
		def __init__(self, signifiers, flags='', length_lower=1, length_upper=None, word_any_length=False, worddict=None):
			# Fix defaults
			# ------------
			if length_upper == None:
				length_upper = lenth_lower

			# Check values
			# ------------
			if not signifiers:
				raise ValueError('Expression must include at least one signifier')
			if len(signifiers) > 1 and 'W' in signifiers and not (flags and '~' in flags):
				raise ValueError('Expression cannot contain multiple signifiers when the `W` signifier is included, unless the `~` flag is used')
			if not (length_lower >= 0 and length_upper > 0 and length_upper >= length_lower):
				raise ValueError('Length values must satisfy: `lower` >= 0 AND `upper` > 0 AND `upper` >= `lower`')

			# Set values
			# ----------
			self.signifiers = {sig for sig in signifiers if sig in Pattern.all_sigs}
			if len(self.signifiers) != len(signifiers):
				raise ValueError('An invalid or duplicate signifier was used. Only one of each of these signifiers are valid: {}'.format(', '.join(['`%s`'%s for s in Pattern.all_sigs])))
			self.flags = {f for f in flags if f in Pattern.all_flags}
			if len(self.flags) != len(flags):
				raise ValueError('An invalid or duplicate flag was used. Only one of each of these flags are valid: {}'.format(', '.join(['`%s`'%f for f in Pattern.all_flags])))
			self.length_lower = length_lower
			self.length_upper = length_upper
			self.word_any_length = word_any_length
			self.worddict = worddict

		def __str__(self):
			out = '%{}{}'.format(''.join(self.signifiers), ''.join(self.flags))
			if len(self.signifiers) > 1:
				out = '{'+out+'}'
			if self.length_lower != 1:
				out += '['+str(self.length_lower)
				if self.length_lower != self.length_upper:
					out += '-'+str(self.length_upper)
				out += ']'
			return out

		def generate(self):
			# Set flag mode variables
			CHOOSE_SIG = '~' in self.flags
			REPEAT_EQ = '=' in self.flags
			CAP_MODE = 0 + (1 if '^' in self.flags else 0) + (2 if '+' in self.flags else 0)
			# Apply the '~' (choose sig) flag
			if CHOOSE_SIG:
				sigs = {random.choice(tuple(self.signifiers))}
			else:
				sigs = self.signifiers
			# Decide whether to use WordGenerator
			if 'W' in sigs:
				if not self.worddict:
					raise ValueError('Attempted to use the `W` signifier while no word dictionary is loaded, load a dictionary with the `-w` or `--worddict` command options')
				if self.word_any_length:
					wordpool = self.worddict.getWordPool()
				else:
					wordpool = self.worddict.getWordPool(self.length_lower, self.length_upper)
				out = random.choice(tuple(wordpool))
			else:
				# Choose length
				length = random.randint(self.length_lower, self.length_upper)
				# Collect pools
				pool = {c for sig in sigs for c in Pattern.pools_dict[sig]}
				# Apply the '=' (repeat same) flag
				if REPEAT_EQ:
					c = random.choice(tuple(pool))
					if CAP_MODE == 1 or CAP_MODE == 2 or CAP_MODE == 3 and random.randrange(2):
						CAP_MODE = 2
					else:
						CAP_MODE = 0
					out = ''.join(repeat(c,length))
				# Generate sequence normally
				else:
					out = ''
					for _ in range(length):
						c = random.choice(tuple(pool))
						out += c
			# Apply the '+' and '^' (capitalization) flags
			if CAP_MODE == 3:
				def randcap(matchobj):
					c = matchobj.group()
					return c.upper() if random.randrange(2) else c
				out = re.sub(r'[a-z]', randcap, out)
			elif CAP_MODE == 2:
				out = out.upper()
			elif CAP_MODE == 1:
				start, c = random.choice([(m.start(), m.group()) for m in re.finditer(r'[a-z]', out)])
				out = out[0:start]+(c.upper())+out[start+1:]
			# Return generated sequence
			return out

	def __init__(self, pattern, worddict=None):
		def compile_expression(match, worddict=None):
			if match:
				gdict = match.groupdict()
				signifiers = gdict['sig'] or gdict['sigm'] or ''
				flags = gdict['flags'] or gdict['flagsm'] or ''
				length_lower = int(gdict['length_lower'] or 1)
				length_upper = int(gdict['length_upper'] or length_lower)
				word_any_length = not bool(gdict['length_lower'])
				try:
					return Pattern.Expression(signifiers, flags, length_lower, length_upper, word_any_length, worddict)
				except ValueError:
					raise ValueError('Failed to compile expression: `{}`'.format(match.group()))
			return None
		# Parse and compile expressions
		self.expressions = []
		pos = 0
		while pos < len(pattern):
			match = Pattern.expression_re.match(pattern, pos)
			exp = compile_expression(match, worddict)
			if not exp:
				raise ValueError('Pattern `{}` failed to compile at index: {}'.format(pattern, pos))
			self.expressions.append(exp)
			pos = match.end()
		# Set word dictionary
		self.worddict = worddict

	def __str__(self):
		return ''.join(str(exp) for exp in self.expressions)

	def generate(self):
		return ''.join(e.generate() for e in self.expressions)


import argparse
import readline
import pyperclip
import sys
from os import path

from .pattern import Pattern
from .worddict import WordDictionary
from .utils import *

# Argument Defaults
# =================
DEFAULT_PATTERN = '%d[4]%s=[2]%W[6-10]'
# Relavent Paths
# ==============
WORDS_FILE = path.join(path.dirname(path.abspath(__file__)), 'words', 'words.txt')

def main():
	args = parser().parse_args()

	# Word Dictionary ops
	if args.revert:
		WordDictionary.revert(WORDS_FILE)
	worddict = None
	if args.worddict:
		print('Generating new words file from file: %r' % args.worddict)
		worddict = WordDictionary.setWordsFile(WORDS_FILE, args.worddict)
	if not worddict:
		try: 
			worddict = WordDictionary(WORDS_FILE)
		except FileNotFoundError:
			worddict = None
			printerr('Could not find words file at %r, can continue as long as pattern does not use the `W` signifier' % WORDS_FILE)

	# Parse pattern
	try:
		pattern = Pattern(args.pattern, worddict)
	except ValueError as e:
		printerr('Error when compiling pattern with `%s`: %s' % (args.pattern, e))
		sys.exit(1)

	# Generate password(s)
	if args.interactive:
		print('Entering interactive mode. Enter anything to generate new password. Enter a new pattern at any time to use instead, if valid.')
		print('  Enter `q` to quit')
		print('  Generating using pattern: `%s`' % pattern)
		quit = False
		while not quit:
			try:
				out = pattern.generate()
			except ValueError as e:
				printerr('Error when generating password: %s' % e)
				sys.exit(1)
			if args.copy:
				pyperclip.copy(out)
				print('  Copying to clipboard')
			print(out)
			if sys.version_info[0] < 3:
				instr = raw_input('> ').strip()
			else:
				instr = input('> ').strip()
			if instr:
				if instr == 'q':
					quit = True
				else:
					try:
						newpattern = Pattern(instr, worddict)
					except ValueError as e:
						print('  Enter `q` to quit')
						printerr('  Error when compiling pattern with `%s`: %s' % (instr, e))
					else:
						pattern = newpattern
						print('  Enter `q` to quit')
						print('  Generating using pattern: `%s`' % pattern)

	else:
		print('  Generating using pattern: `%s`' % pattern)
		out = pattern.generate()
		if args.copy:
			pyperclip.copy(out)
			print('  Copying to clipboard')
		print(out)
	sys.exit(0)
		

def parser():
	parser = argparse.ArgumentParser(	description='Generate random passwords using a pattern ot specify the general format.',
										epilog=howto(),
										formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument(	'pattern',
							nargs='?',
							default=DEFAULT_PATTERN,
							help='The pattern to use to generate the password (defaults to `%(default)s`)')
	parser.add_argument(	'-c', '--copy',
							action='store_true',
							help='Whenever a password is succesfully generated (in either singlue-use mode or interactive mode), '
								+'the string will be copied to your clipboard (may require external libraries, depending on platform')
	parser.add_argument(	'-i', '--interactive',
							action='store_true',
							help='Launches in interactive mode, where passwords of the given pattern are continuously printed after each input, '
								+'and if a valid pattern is given as input at any time, then the new pattern will be used going forward (enter `q` to exit)')
	parser.add_argument(	'-w', '--worddict',
							type=str,
							nargs='?',
							help='Sets the `words.txt` file that iss used as the dictionary for the generator when generating whole words. '
								+'The parser goes line by line, using non-word characters to separate each word (this excludes hyphens and apostrophes, '
								+'which are removed prior to parsing and the two sides of the word are merged) and a new, formatted `words.txt` '
								+'file will be created (the previous version will be copied to words.txt.old)')
	parser.add_argument(	'-R', '--revert',
							action='store_true',
							help='Reverts the worddict file at `words.txt` with the backup file, if there is one. '
							+'This is performed before a new `words.txt` file is generated if the `-w` command is used with this')
	return parser	

def howto():
	return 'Go to `https://github.com/nkrim/passwordgen` to see the README for the how-to documentation\n'

if __name__ == '__main__':
	main()


import os
import re
from shutil import copyfileobj
from tempfile import TemporaryFile, mkstemp

from .utils import *

class WordDictionary:
	class LengthSetMap:
		def __init__(self):
			self._words = [{''}]

		def __bool__(self):
			return self.maxlength() > 0

		def __getitem__(self, length):
			return self._words[length]

		def __iter__(self):
			return self._words.__iter__()

		def __len__(self):
			return len(self._words)

		def __str__(self):
			return '\n'.join(','.join(sorted(word_set)) for word_set in self._words[1:self.maxlength()+1])

		def add(self, word, length=-1):
			if length < 0:
				length = len(word)
			if length+1 > len(self._words):
				self._words += [set()] * (length+1-len(self._words))
			self._words[length].add(word)

		def maxlength(self):
			for i in reversed(range(len(self._words))):
				if len(self._words[i]) > 0:
					return i
			return 0


	def __init__(self, words_file, wordmap=None):
		self.words_file = words_file
		if not wordmap:
			wordmap = WordDictionary.parse(self.words_file, formatted=True)
		self.wordmap = wordmap

	def getWordPool(self, length_lower=None, length_upper=None):
		if not length_upper:
			length_upper = length_lower

		if length_lower != None:
			pool = {w for lenset in self.wordmap[length_lower:length_upper+1] for w in lenset}
		else:
			pool = set()

		if not pool:
			pool = {w for lenset in self.wordmap[1:] for w in lenset}
			if not pool:
				pool = {w for w in self.wordmap[0]}
		return pool

	@staticmethod
	def parse(file_path, formatted=False):
		wordmap = WordDictionary.LengthSetMap()
		if formatted:
			length = 1
			with open(file_path, 'r') as f:
				for line in f:
					for w in line.split(','):
						if w:
							wordmap.add(w, length)
					length += 1
		else:
			sub_re = re.compile(r'[\-\']')
			split_re = re.compile(r'\W+')
			with open(file_path, 'r') as f:
				for line in f:
					for w in split_re.split(sub_re.sub('', line)):
						if w:
							wordmap.add(w)
		return wordmap

	@staticmethod
	def backup(words_file):
		# Copy old `words.txt` to `words.txt.old`
		try:
			with open(words_file, 'r') as f:
				with open(words_file+'.old', 'w') as old:
					copyfileobj(f, old)
		except IOError:
			printerr('No formatted words file could be found at %r, skipping backup' % words_file)
		except:
			printerr('Could not backup words file from %r to %r' % (words_file, words_file+'.old'))
		else:
			return True
		return False

	@staticmethod
	def revert(words_file):
		# Revert `words.txt.old` to `words.txt`
		_, temp_file = mkstemp()
		old_file = words_file+'.old'
		try: 
			with open(old_file, 'r') as old:
				with open(temp_file, 'w') as temp:
					copyfileobj(old, temp)
		except IOError:
			printerr('No backup file found at %r' % old_file)
		except:
			printerr('Could not load backup file %r' % old_file)
		else:
			if WordDictionary.backup(words_file):
				try:
					with open(temp_file, 'r') as temp:
						with open(words_file, 'w') as f:
							copyfileobj(temp, f)
				except IOError:
					printerr('No words file found at %r' % words_file)
				except:
					printerr('Could not revert backup to %r, attempting to restore overwritten backup' % words_file)
					try: 
						with open(temp_file, 'r') as temp:
							with open(old_file, 'w') as old:
								copyfileobj(temp, old)
					except:
						printerr('Could not restore the overwritten backup. Backup is lost.')
				else:
					os.remove(temp_file)
					return True
		os.remove(temp_file)
		return False

	@staticmethod
	def setWordsFile(words_file, file_path, backup=True):
		# Read input file
		try:
			wordmap = WordDictionary.parse(file_path)
		except FileNotFoundError:
			printerr('Could not find file %r' % file_path)
			return None
		# Backup words file
		if backup:
			WordDictionary.backup(words_file)
		# Write new words file
		try: 
			with open(words_file, 'w') as f:
				f.write(str(wordmap))
		except Exception as e:
			printerr('Could not write new words file: %s' % e)
			return None
		# Return wordmap
		return WordDictionary(words_file, wordmap)


import setuptools
from distutils.core import setup
from os import path

# Helper functions
# ----------------
def readFile(file):
	with open(file) as f:
		return f.read()

# Arguments
# ------------
CLASSIFIERS = []

# Dynamic info
# ------------
VERSION 			= '0.1.5'
CLASSIFIERS 		+= [
						'Development Status :: 3 - Alpha',
					]

# Package/Dependency info
# ---------------
PACKAGES			= [	'passwordgen'	]
PACKAGE_DIR 		= {	'passwordgen': 'src'	}
PACKAGE_DATA 		= {	'passwordgen': ['words/words.txt', 'words/defaults/*.txt']	}
DATA_FILES			= [ ('', ['README.rst','LICENSE']), ]
INSTALL_REQUIRES 	= [	'pyperclip>=1.5.27' ]

# Static info
# -----------
NAME 				= 'passwordgen'
DESCRIPTION 		= 'A generator for safe and random passwords defined by a user-defined pattern'
LONG_DESCRIPTION 	= readFile(path.join(path.dirname(path.abspath(__file__)), 'README.rst'))
AUTHOR 				= 'Noah Krim'
AUTHOR_EMAIL 		= 'nkrim62@gmail.com'
LICENSE 			= 'MIT License'
URL 				= 'https://github.com/nkrim/passwordgen'
KEYWORDS 			= 'passwordgen password generator safe random pattern'
ENTRY_POINTS		= { 'console_scripts': [ 'passwordgen = passwordgen.__main__:main' ] }
CLASSIFIERS 		+= [
						'Environment :: Console',
						'Intended Audience :: End Users/Desktop',
						'License :: OSI Approved :: MIT License',
						'Natural Language :: English',
						'Operating System :: OS Independent',
						'Programming Language :: Python',
						'Programming Language :: Python :: 2',
						'Programming Language :: Python :: 3',
						'Topic :: Security',
						'Topic :: Utilities',
					]
INC_PKG_DATA		= True
ZIP_SAFE			= False

# Setup call
# ----------
setup(
	name=NAME,
	version=VERSION,
	description=DESCRIPTION,
	long_description=LONG_DESCRIPTION,
	author=AUTHOR,
	author_email=AUTHOR_EMAIL,
	license=LICENSE,
	url=URL,
	keywords=KEYWORDS,
	entry_points=ENTRY_POINTS,
	packages=PACKAGES,
	package_dir=PACKAGE_DIR,
	package_data=PACKAGE_DATA,
	data_files=DATA_FILES,
	install_requires=INSTALL_REQUIRES,
	classifiers=CLASSIFIERS,
	include_package_data=INC_PKG_DATA,
	zip_safe=ZIP_SAFE )